from __future__ import absolute_import
from __future__ import unicode_literals

from collections import OrderedDict
import os

import requests
from git import Repo
from slugify import slugify

from accountable.config import Config


class Accountable(object):
    def __init__(self):
        self.issue_key = None

    def projects(self):
        metadata = self._metadata()
        return [(project['id'],
                 project['key'],
                 project['name']) for project in metadata['projects']]

    @property
    def resource(self):
        return Resource()

    def issue_types(self, project_key):
        metadata = self._metadata()
        if project_key:
            project = next(
                project for project in metadata['projects']
                if project['key'] == str(project_key)
            )
            return {project_key: project['issuetypes']}
        else:
            issue_types = {}
            for project in metadata['projects']:
                issue_types[project['key']] = project['issuetypes']
            return issue_types

    def issue(self):
        return self.resource.get('issue/{}'.format(self.issue_key))

    def issue_meta(self):
        fields = self.issue()['fields']

        data = OrderedDict()
        for field in Config()['issue_fields']:
            field_name = self._field_name(field)
            data[field_name] = self._access_field(field, fields)
        return data

    def issue_create(self, options):
        payload = self._args_to_dict(options)
        return self.resource.post('issue', payload)

    def checkout_branch(self, options):
        payload = self._args_to_dict(options)
        new_issue = self.issue_create(options)
        summary = payload['fields']['summary']
        key = new_issue['key']
        self._repo().checkout('HEAD', b='{}-{}'.format(key, slugify(summary)))
        return new_issue

    def checkout(self, issue_key):
        self.issue_key = issue_key
        issue = self.issue()
        issue_data = {
            'self': issue['self'],
            'id': issue['id'],
            'key': issue['key']
        }
        branch_name = '{}-{}'.format(issue_key, slugify(
            issue['fields']['summary'])
        )

        self._repo().branch('-f', branch_name)
        self._repo().checkout(branch_name)
        return issue_data

    def issue_comments(self):
        return self.resource.get('issue/{}/comment'.format(self.issue_key))

    def issue_add_comment(self, body):
        return self.resource.post('issue/{}/comment'.format(self.issue_key),
                                  {'body': body})

    def issue_worklog(self):
        return self.resource.get('issue/{}/worklog'.format(self.issue_key))

    def issue_transitions(self):
        return self.resource.get(
            'issue/{}/transitions'.format(self.issue_key)
        )

    def issue_do_transition(self, transition_id):
        return self.resource.post(
            'issue/{}/transitions'.format(self.issue_key),
            {'transition': {'id': transition_id}}
        )

    def users(self, query):
        payload = self.resource.get('user/search',
                                    params={'username': query})
        return payload

    def _repo(self):
        return Repo(os.getcwd()).git

    def _metadata(self):
        return self.resource.get('issue/createmeta')

    def _args_to_dict(self, args_tuple):
        d = {}
        for arg in zip(args_tuple[0::2], args_tuple[1::2]):
            keys = arg[0].split('.')
            self._set_nested_key(keys, arg[1], d)
        return {'fields': d}

    def _set_nested_key(self, key, value, d):
        if isinstance(key, list) and len(key) > 1:
            head, tail = key[0], key[1:]
            if not d.get(head):
                d[head] = {}
            self._set_nested_key(tail, value, d[head])
        else:
            d[list(key)[0]] = value

        return d

    @staticmethod
    def _access_field(field, d):
        if isinstance(field, str):
            return d[field]
        elif isinstance(field, dict):
            value = d[list(field.keys())[0]]
            return Accountable._access_field(list(field.values())[0], value)
        else:
            raise TypeError('There is an issue with your issue field'
                            'configuration.')

    @staticmethod
    def _field_name(field):
        if isinstance(field, str):
            return field.upper()
        else:
            return list(field.keys())[0].upper()


class Resource(object):
    def __init__(self):
        self.config = Config()
        self.auth = self.config.auth
        self.api_uri = 'https://{}/rest/api/2'.format(self.config['domain'])

    def get(self, resource, params={}):
        r = requests.get('{}/{}'.format(self.api_uri, resource),
                         params=params,
                         auth=self.auth)
        return r.json()

    def post(self, resource, payload={}):
        r = requests.post('{}/{}'.format(self.api_uri, resource),
                          auth=self.auth,
                          json=payload)
        try:
            return r.json()
        except ValueError:
            return r



import os


import click
import yaml

from requests.auth import HTTPBasicAuth


class Config(object):
    CONFIG_DIR = os.path.expanduser('~/.accountable')
    DEFAULT_ISSUE_FIELDS = [
        {'reporter': 'displayName'},
        {'assignee': 'displayName'},
        {'issuetype': 'name'},
        {'status': {'statusCategory': 'name'}},
        'summary',
        'description'
    ]
    DEFAULT_ALIASES = {'cob': 'checkoutbranch',
                       'co': 'checkout'}

    def __init__(self):
        self._config = None

    def __getitem__(self, name):
        return self.config[name]

    def __repr__(self):
        '{}({})'.format(self.__class__, self.config)

    def create(self, **kwargs):
        username = kwargs.get('username')
        password = kwargs.get('password')
        domain = kwargs.get('domain')
        config_dict = self._config_dict(username, password, domain)
        self._create_config(config_dict)

    @property
    def config(self):
        if self._config is None:
            self._config = self._load_config()
        return self._config

    @property
    def auth(self):
        return HTTPBasicAuth(self['username'], self['password'])

    @property
    def config_file(self):
        return '{}/config.yaml'.format(self.CONFIG_DIR)

    def _load_config(self):
        with open(self.config_file, 'r') as f:
            config = yaml.load(f)
        return config

    def _create_config_dir(self):
        if not os.path.exists(self.CONFIG_DIR):
            click.echo('Creating {}'.format(self.CONFIG_DIR))
            os.makedirs(self.CONFIG_DIR)

    def _config_dict(self, username, password, domain):
        return {
            'username': username,
            'password': password,
            'domain': domain,
            'issue_fields': self.DEFAULT_ISSUE_FIELDS,
            'aliases': self.DEFAULT_ALIASES
        }

    def _create_config(self, config_dict):
        self._create_config_dir()
        with open(self.config_file, 'w+') as f:
            f.write(yaml.dump(config_dict, default_flow_style=False,
                              encoding='utf-8'))
        click.echo('Configuration file written to {}'.format(self.config_file))



from __future__ import absolute_import

from functools import update_wrapper

import click

from accountable.accountable import Accountable, Config


pass_accountable = click.make_pass_decorator(Accountable)


class AccountableCli(click.Group):
    def get_command(self, ctx, cmd_name):
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv
        alias = Config()['aliases'].get(cmd_name, None)
        if alias is not None:
            return click.Group.get_command(self, ctx, alias)
        return None


def prettyprint(output):
    if isinstance(output, list):
        for i in output:
            click.echo(' - '.join([x for x in i]))
    else:
        click.echo(' - '.join([x for x in output]))


@click.group(cls=AccountableCli)
@click.pass_context
def cli(ctx):
    """
    A Jira CLI.
    """
    ctx.obj = Accountable()


@click.command()
@click.option('--username', prompt='Your Jira username')
@click.option('--password', prompt='Your Jira password', hide_input=True)
@click.option('--domain', prompt='The domain where your account is hosted')
def configure(username, password, domain):
    """
    Initial configuration. Used to specify your username, password and domain.
    Configuration is stored in ~/.accountable/config.yaml.
    """
    Config().create(username=username, password=password, domain=domain)


@click.command()
@pass_accountable
def projects(accountable):
    """
    List all projects.
    """
    projects = accountable.projects()
    for pid, key, name in projects:
        prettyprint((pid, key, name))


@click.command()
@click.argument('project_key', default='')
@pass_accountable
def issuetypes(accountable, project_key):
    """
    List all issue types. Optional parameter to list issue types by a given
    project.
    """
    projects = accountable.issue_types(project_key)
    for key, issue_types in sorted(projects.items()):
        for i in issue_types:
            prettyprint((i['id'], key, i['name'], i['description']))


def nargs(f):
    @click.argument('options', nargs=-1)
    @pass_accountable
    @click.pass_context
    def new_func(ctx, accountable, options, *args, **kwargs):
        return ctx.invoke(f, accountable, options, *args, **kwargs)
    return update_wrapper(new_func, f)


@click.command()
@nargs
def createissue(accountable, options):
    """
    Create new issue.
    """
    issue = accountable.issue_create(options)
    prettyprint((issue['id'], issue['key'], issue['self']))


@click.command()
@nargs
def checkoutbranch(accountable, options):
    """
    Create a new issue and checkout a branch named after it.
    """
    issue = accountable.checkout_branch(options)
    prettyprint((issue['id'], issue['key'], issue['self']))


@click.command()
@click.argument('issue_key')
@pass_accountable
def checkout(accountable, issue_key):
    """
    Checkout a new branch or checkout to a branch for a given issue.
    """
    issue = accountable.checkout(issue_key)
    prettyprint((issue['id'], issue['key'], issue['self']))


@click.group(invoke_without_command=True)
@click.argument('issue_key')
@pass_accountable
@click.pass_context
def issue(ctx, accountable, issue_key):
    """
    List metadata for a given issue key. Issue keys should take the format of
    {PROJECT-ID}-{ISSUE-ID}.
    """
    accountable.issue_key = issue_key
    if not ctx.invoked_subcommand:
        issue = accountable.issue_meta()
        for field, value in issue.items():
            prettyprint((field, value))


@click.command()
@pass_accountable
def comments(accountable):
    """
    Lists all comments for a given issue key.
    """
    comments = accountable.issue_comments().get('comments')
    if comments:
        for c in comments:
            prettyprint((c['id'], c['author']['name'], c['body'],
                         c['created']))
    else:
        prettyprint(('No comments found for {}'.format(
            accountable.issue_key), )
        )


@click.command()
@click.argument('body')
@pass_accountable
def addcomment(accountable, body):
    """
    Add a comment to the given issue key. Accepts a body argument to be used
    as the comment's body.
    """
    r = accountable.issue_add_comment(body)
    prettyprint((r['author']['name'], r['body'], r['created']))


@click.command()
@pass_accountable
def worklog(accountable):
    """
    List all worklogs for a given issue key.
    """
    worklog = accountable.issue_worklog().get('worklogs')
    if worklog:
        for w in worklog:
            prettyprint(('Author', w['author']['name']))
            prettyprint(('Comment', w.get('comment')))
            prettyprint(('Time spent', w['timeSpent']))
    else:
        prettyprint(('No worklogs found for {}'.format(accountable.issue_key),
                     ))


@click.command()
@pass_accountable
def transitions(accountable):
    """
    List all possible transitions for a given issue.
    """
    transitions = accountable.issue_transitions().get('transitions')
    if transitions:
        for t in transitions:
            prettyprint((t['id'], t['name']))
    else:
        prettyprint(('No transitions found for {}'.format(
            accountable.issue_key),
        ))


@click.command()
@pass_accountable
@click.argument('transition_id')
def dotransition(accountable, transition_id):
    """
    Transition the given issue to the provided ID. The API does not return a
    JSON response for this call.
    """
    t = accountable.issue_do_transition(transition_id)
    if t.status_code == 204:
        prettyprint((
            'Successfully transitioned {}'.format(accountable.issue_key),
        ))


@click.command()
@pass_accountable
@click.argument('query')
def users(accountable, query):
    """
    Executes a user search for the given query.
    """
    users = accountable.users(query)
    prettyprint([(user['key'], user['displayName']) for user in users])


issue.add_command(dotransition)
issue.add_command(transitions)
issue.add_command(worklog)
issue.add_command(addcomment)
issue.add_command(comments)

cli.add_command(configure)
cli.add_command(projects)
cli.add_command(issuetypes)
cli.add_command(issue)
cli.add_command(createissue)
cli.add_command(users)
cli.add_command(checkoutbranch)
cli.add_command(checkout)


if __name__ == '__main__':
    cli()



from setuptools import setup, find_packages
from pip.req import parse_requirements


with open('README.rst') as f:
    long_description = f.read()

install_reqs = parse_requirements('requirements.txt', session=False)
reqs = [str(ir.req) for ir in install_reqs]

setup(
    name='accountable',
    version='0.1',
    description='Command line tools for interacting with JIRA',
    long_description=long_description,
    url='https://github.com/wohlgejm/accountable',
    author='Jerry Wohlgemuth',
    author_email='wohlgejm@gmail.com',
    license='MIT',
    keywords='jira',
    packages=find_packages(exclude=['docs', 'tests']),
    install_requires=reqs,
    entry_points='''
        [console_scripts]
        accountable=accountable.cli:cli
    ''',
)
